/*
 * Copyright (c) 2018, UAB 8Devices. All rights reserved.
 */

#include "8192cd_cfg.h"
#include "8192cd.h"
#include "8192cd_headers.h"
#include "8192cd_debug.h"
#include "awds.h"

#define QUALCOMM_OUI				0xf0fd8c
#define QUALCOMM_OUI_WHC_TYPE			0x00
#define QUALCOMM_OUI_WHC_AP_INFO_SUBTYPE	0x00
#define QUALCOMM_OUI_WHC_AP_INFO_VERSION1	0x01
#define QUALCOMM_OUI_WHC_AP_INFO_CAP_WDS	0x01

#define awds_printk(lvl, priv, fmt, args...) \
	printk(lvl "[%s]" fmt, (priv)->dev->name, ##args)
#define awds_info(priv, fmt, args...) \
	awds_printk(KERN_INFO, (priv), fmt, ##args)

static void rtl8192cd_awds_notify(void *src, const u8 *addr,
				  enum awds_fdb_action action, void *data);

int rtl8192cd_init_awds(struct rtl8192cd_priv *priv)
{
	priv->awds_fdb = awds_fdb_create();
	if (priv->awds_fdb == NULL)
		return -1;

	awds_fdb_set_notify(priv->awds_fdb, rtl8192cd_awds_notify, priv);

	return 0;
}

void rtl8192cd_deinit_awds(struct rtl8192cd_priv *priv)
{
	if (priv->awds_fdb == NULL)
		return;

	awds_fdb_release(priv->awds_fdb);
	priv->awds_fdb = NULL;
}

void rtl8192cd_open_awds(struct rtl8192cd_priv *priv)
{
	awds_fdb_enable_gc(priv->awds_fdb);
}

void rtl8192cd_close_awds(struct rtl8192cd_priv *priv)
{
	awds_fdb_disable_gc(priv->awds_fdb);
}

struct stat_info *rtl8192cd_get_stainfo_awds(struct rtl8192cd_priv *priv,
					     const unsigned char *hwaddr)
{
	return (struct stat_info *)awds_fdb_lookup(priv->awds_fdb, hwaddr);
}

void rtl8192cd_release_stainfo_awds(struct rtl8192cd_priv *priv,
				    struct stat_info *pstat)
{
	awds_fdb_flush(priv->awds_fdb, pstat);
}

void rtl8192cd_stat_awds_enable(struct rtl8192cd_priv *priv,
				struct stat_info *pstat)
{
	if (!RTL8192CD_STAT_IS_AWDS_ENABLED(priv, pstat)) {
		RTL8192CD_STAT_AWDS_ENABLE(priv, pstat);
		awds_info(priv, "[%pM]: AWDS enabled\n", pstat->hwaddr);
	}
}

int rtl8192cd_is_awds_ap(struct rtl8192cd_priv *priv, struct rx_frinfo *pfrinfo)
{
	unsigned char *pframe = get_pframe(pfrinfo);
	unsigned char *p = pframe + WLAN_HDR_A3_LEN + _ASOCRSP_IE_OFFSET_;
	int len = 0;

	BUG_ON((WIFI_ASSOCRSP != GetFrameSubType(pframe)) &&
	       (WIFI_REASSOCRSP != GetFrameSubType(pframe)));

	while ((p = get_ie(p, _RSN_IE_1_, &len,
			   pfrinfo->pktlen - (p - pframe)))) {
		u32 oui = get_unaligned_le32(p + 2);

		if ((oui == (QUALCOMM_OUI_WHC_TYPE << 24 | QUALCOMM_OUI)) &&
		    (*(p + 6) == QUALCOMM_OUI_WHC_AP_INFO_SUBTYPE)) {
			/* Qualcomm's Whole Home Coverage AP */
			if ((*(p + 7) == QUALCOMM_OUI_WHC_AP_INFO_VERSION1) &&
			    (get_unaligned_le16(p + 8) &
			     QUALCOMM_OUI_WHC_AP_INFO_CAP_WDS)) {
				return 1;
			}
		}

		p += len + 2;
	}

	return 0;
}

static void rtl8192cd_awds_notify(void *src, const u8 *addr,
				  enum awds_fdb_action action, void *data)
{
	struct rtl8192cd_priv *priv = (struct rtl8192cd_priv *)data;
	struct stat_info *pstat = (struct stat_info *)src;

	switch (action) {
	case AWDS_FDB_ADD:
	case AWDS_FDB_CHANGE:
		rtl8192cd_stat_awds_enable(priv, pstat);
		awds_info(priv, "[%pM]: new forward address %pM\n",
			  pstat->hwaddr, addr);
		break;

	case AWDS_FDB_REMOVE:
		awds_info(priv, "[%pM]: removed forward address %pM\n",
			  pstat->hwaddr, addr);
		break;
	}
}

static void rtl8192cd_update_awds_fdb(struct rtl8192cd_priv *priv,
				      struct stat_info *pstat, const u8 *addr)
{
	awds_fdb_update(priv->awds_fdb, pstat, addr);
}

#ifdef MBSSID
int rtl8192cd_rx_dispatch_awds(struct rtl8192cd_priv *priv,
			       struct rx_frinfo *pfrinfo, int vap_idx)
#else
int rtl8192cd_rx_dispatch_awds(struct rtl8192cd_priv *priv,
			       struct rx_frinfo *pfrinfo)
#endif
{
	unsigned char *pframe = get_pframe(pfrinfo);
	unsigned int frtype = GetFrameType(pframe);
	unsigned int opmode;
	unsigned char *myhwaddr;
	struct stat_info *pstat;

	priv = GET_PRIV(priv, vap_idx);
	opmode = OPMODE;
	myhwaddr = GET_MY_HWADDR;

	rx_sum_up(priv, NULL, pfrinfo);

	if (frtype != WIFI_DATA_TYPE)
		return 1;
	if (!IS_MCAST(GetAddr1Ptr(pframe)) &&
	    !isEqualMACAddr(myhwaddr, GetAddr1Ptr(pframe)))
		return 1;
	if (isEqualMACAddr(myhwaddr, GetAddr2Ptr(pframe)) ||
	    isEqualMACAddr(myhwaddr, GetAddr4Ptr(pframe)))
		return 1;

#ifdef CLIENT_MODE
	if ((opmode & (WIFI_STATION_STATE | WIFI_ASOC_STATE)) ==
	    (WIFI_STATION_STATE | WIFI_ASOC_STATE)) {
		if (!IS_BSSID(priv, GetAddr2Ptr(pframe))) {
			return 1;
		}
	} else
#endif
		if (opmode & WIFI_AP_STATE) {
		if (IS_MCAST(GetAddr1Ptr(pframe))) {
			return 1;
		}
	} else {
		return 1;
	}

	pstat = get_stainfo(priv, GetAddr2Ptr(pframe));
	if (!pstat)
		return 1;

	rtl8192cd_update_awds_fdb(priv, pstat, pfrinfo->sa);
#ifdef RTL8190_DIRECT_RX
	rtl8192cd_rx_dataframe(priv, NULL, pfrinfo);
#else
	SMP_LOCK_RX_DATA(x);
	list_add_tail(&(pfrinfo->rx_list), &(priv->rx_datalist));
	SMP_UNLOCK_RX_DATA(x);
#endif

	return 0;
}

int rtl8192cd_xmit_awds_mcast(struct rtl8192cd_priv *priv, struct sk_buff *skb)
{
	struct list_head *phead, *plist;
	struct stat_info *pstat, *ta_pstat;
	struct sk_buff *newskb;
	unsigned char *sa = skb->data + MACADDRLEN;
	int sta_count = 0;

	ta_pstat = get_stainfo(priv, sa);
	if (!ta_pstat)
		ta_pstat = rtl8192cd_get_stainfo_awds(priv, sa);

	phead = &priv->asoc_list;
	plist = phead->next;
	while (phead && (plist != phead)) {
		pstat = list_entry(plist, struct stat_info, asoc_list);
		plist = plist->next;

		if (pstat == ta_pstat)
			continue;

		if (!(pstat->state & WIFI_AWDS)) {
			sta_count++;
			continue;
		}

		newskb = skb_copy(skb, GFP_ATOMIC);
		if (!newskb)
			return 1;

		newskb->cb[2] = (char)0xff; /* do not aggregate */
		__rtl8192cd_start_xmit_out(newskb, pstat);
	}

	return !sta_count;
}

void rtl8192cd_xmit_awds_ucast_unknown(struct rtl8192cd_priv *priv,
				       struct sk_buff *skb)
{
	rtl8192cd_xmit_awds_mcast(priv, skb);
}

void issue_4addr_NullData(struct rtl8192cd_priv *priv, unsigned char *da)
{
	struct wifi_mib *pmib = GET_MIB(priv);
	unsigned char *myhwaddr = GET_MY_HWADDR;
	DECLARE_TXINSN(txinsn);

	txinsn.q_num = MANAGE_QUE_NUM;
	txinsn.pframe = NULL;
	txinsn.phdr = get_wlanhdr_from_poll(priv);
	txinsn.hdr_len = WLAN_HDR_A4_LEN;
	txinsn.fr_type = _PRE_ALLOCHDR_;
#ifdef P2P_SUPPORT
	if (rtk_p2p_is_enabled(priv))
		txinsn.tx_rate = _6M_RATE_;
	else
#endif
		txinsn.tx_rate = find_rate(priv, NULL, 0, 1);
#ifndef TX_LOWESTRATE
	txinsn.lowest_tx_rate = txinsn.tx_rate;
#endif
	txinsn.fixed_rate = 1;
	txinsn.retry = pmib->dot11OperationEntry.dot11ShortRetryLimit;

	if (txinsn.phdr == NULL)
		goto err_exit;

	memset((void *)(txinsn.phdr), 0, sizeof(struct wlan_hdr));
	SetFrameSubType(txinsn.phdr, WIFI_DATA_NULL);
	SetToDs(txinsn.phdr);
	SetFrDs(txinsn.phdr);
	memcpy((void *)GetAddr1Ptr(txinsn.phdr), da, MACADDRLEN);
	memcpy((void *)GetAddr2Ptr(txinsn.phdr), myhwaddr, MACADDRLEN);
	memcpy((void *)GetAddr3Ptr(txinsn.phdr), da, MACADDRLEN);
	memcpy((void *)GetAddr4Ptr(txinsn.phdr), myhwaddr, MACADDRLEN);

	if ((rtl8192cd_firetx(priv, &txinsn)) == SUCCESS)
		return;

err_exit:
	if (txinsn.phdr)
		release_wlanhdr_to_poll(priv, txinsn.phdr);
}

int MDL_INIT rtl8192cd_awds_init(void)
{
	int err;

	err = awds_init();
	if (err)
		return err;

	return 0;
}

void MDL_EXIT rtl8192cd_awds_exit(void)
{
	awds_exit();
}
