/*
 * Copyright (c) 2018, UAB 8Devices. All rights reserved.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/jhash.h>
#include <linux/if_ether.h>
#include <linux/etherdevice.h>
#include <linux/hashtable.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include "awds.h"

struct awds_fdb_entry {
	struct hlist_node	node;
	void			*dst;
	u8			addr[ETH_ALEN];

	unsigned long		updated ____cacheline_aligned_in_smp;
	struct rcu_head		rcu;
};

#define AWDS_FDB_DEFAULT_AGEING_TIME	(300 * HZ)

struct awds_fdb {
	spinlock_t		lock;
	DECLARE_HASHTABLE	(fdb_hash, 6);
	void			(*notify)(void *src, const u8 *addr,
					  enum awds_fdb_action action,
					  void *data);
	void			*notify_data;
	unsigned long		ageing_time;
	struct delayed_work	gc_work;
};

static struct kmem_cache *awds_fdb_cache __read_mostly;
static void awds_fdb_delete_all(struct awds_fdb *fdb);
static void awds_fdb_gc(struct work_struct *work);

static void awds_fdb_init(struct awds_fdb *fdb)
{
	spin_lock_init(&fdb->lock);
	hash_init(fdb->fdb_hash);
	fdb->ageing_time = AWDS_FDB_DEFAULT_AGEING_TIME;
	INIT_DELAYED_WORK(&fdb->gc_work, awds_fdb_gc);
}

static void awds_fdb_cleanup(struct awds_fdb *fdb)
{
	cancel_delayed_work_sync(&fdb->gc_work);
	awds_fdb_delete_all(fdb);
}

struct awds_fdb *awds_fdb_create(void)
{
	struct awds_fdb *fdb;

	fdb = kzalloc(sizeof(*fdb), GFP_KERNEL);
	if (!fdb)
		return NULL;

	awds_fdb_init(fdb);

	return fdb;
}

void awds_fdb_release(struct awds_fdb *fdb)
{
	awds_fdb_cleanup(fdb);
	kfree(fdb);
}

void awds_fdb_set_notify(struct awds_fdb *fdb,
			 void (*notify)(void *src, const u8 *addr,
					enum awds_fdb_action action,
					void *data),
			 void *data)
{
	fdb->notify_data = data;
	fdb->notify = notify;
}

static void awds_fdb_notify(struct awds_fdb *fdb, struct awds_fdb_entry *entry,
			    enum awds_fdb_action action)
{
	if (fdb->notify)
		fdb->notify(entry->dst, entry->addr, action, fdb->notify_data);
}

void awds_fdb_enable_gc(struct awds_fdb *fdb)
{
	mod_delayed_work(system_long_wq, &fdb->gc_work, HZ / 10);
}

void awds_fdb_disable_gc(struct awds_fdb *fdb)
{
	cancel_delayed_work_sync(&fdb->gc_work);
}

static struct awds_fdb_entry *awds_fdb_create_entry(struct awds_fdb *fdb,
						    void *src, const u8 *addr)
{
	struct awds_fdb_entry *entry;

	entry = kmem_cache_alloc(awds_fdb_cache, GFP_ATOMIC);
	if (entry) {
		memcpy(entry->addr, addr, ETH_ALEN);
		entry->dst = src;
		entry->updated = jiffies;
		hash_add_rcu(fdb->fdb_hash, &entry->node,
			     jhash(addr, ETH_ALEN, 0));
	}

	return entry;
}

static void awds_fdb_free_entry_rcu(struct rcu_head *head)
{
	struct awds_fdb_entry *entry
		= container_of(head, struct awds_fdb_entry, rcu);

	kmem_cache_free(awds_fdb_cache, entry);
}

static void awds_fdb_delete_entry(struct awds_fdb_entry *entry)
{
	hash_del_rcu(&entry->node);
	call_rcu(&entry->rcu, awds_fdb_free_entry_rcu);
}

static void awds_fdb_delete_all(struct awds_fdb *fdb)
{
	struct awds_fdb_entry *entry;
	int i;

	rcu_read_lock();
	spin_lock(&fdb->lock);

	hash_for_each_rcu(fdb->fdb_hash, i, entry, node)
		awds_fdb_delete_entry(entry);

	spin_unlock(&fdb->lock);
	rcu_read_unlock();
}

static struct awds_fdb_entry *awds_fdb_lookup_rcu(struct awds_fdb *fdb, const u8 *addr)
{
	struct awds_fdb_entry *entry;

	hash_for_each_possible_rcu(fdb->fdb_hash, entry, node,
				   jhash(addr, ETH_ALEN, 0)) {
		if (ether_addr_equal(entry->addr, addr))
			return entry;
	}

	return NULL;
}

void *awds_fdb_lookup(struct awds_fdb *fdb, const u8 *addr)
{
	struct awds_fdb_entry *entry;
	void *dst = NULL;

	rcu_read_lock();

	entry = awds_fdb_lookup_rcu(fdb, addr);
	if (entry)
		dst = entry->dst;

	rcu_read_unlock();

	return dst;
}

void awds_fdb_update(struct awds_fdb *fdb, void *src, const u8 *addr)
{
	struct awds_fdb_entry *entry;

	rcu_read_lock();

	entry = awds_fdb_lookup_rcu(fdb, addr);
	if (likely(entry)) {
		unsigned long now = jiffies;

		if (entry->updated != now)
			entry->updated = now;
		if (unlikely(entry->dst != src)) {
			entry->dst = src;
			awds_fdb_notify(fdb, entry, AWDS_FDB_CHANGE);
		}
	} else {
		spin_lock(&fdb->lock);
		entry = awds_fdb_create_entry(fdb, src, addr);
		if (entry)
			awds_fdb_notify(fdb, entry, AWDS_FDB_ADD);

		spin_unlock(&fdb->lock);
	}

	rcu_read_unlock();
}

void awds_fdb_flush(struct awds_fdb *fdb, void *src)
{
	struct awds_fdb_entry *entry;
	int i;

	rcu_read_lock();
	spin_lock(&fdb->lock);

	hash_for_each_rcu(fdb->fdb_hash, i, entry, node) {
		if (entry->dst == src) {
			awds_fdb_notify(fdb, entry, AWDS_FDB_REMOVE);
			awds_fdb_delete_entry(entry);
		}
	}

	spin_unlock(&fdb->lock);
	rcu_read_unlock();
}

static void awds_fdb_gc(struct work_struct *work)
{
	struct awds_fdb *fdb = container_of(work, struct awds_fdb,
					    gc_work.work);
	struct awds_fdb_entry *entry;
	unsigned long delay = fdb->ageing_time;
	unsigned long work_delay = delay;
	unsigned long now = jiffies;
	int i;

	rcu_read_lock();
	spin_lock(&fdb->lock);

	hash_for_each_rcu(fdb->fdb_hash, i, entry, node) {
		unsigned long expire = entry->updated + delay;

		if (time_after(expire, now)) {
			work_delay = min(work_delay, expire - now);
		} else {
			awds_fdb_notify(fdb, entry, AWDS_FDB_REMOVE);
			awds_fdb_delete_entry(entry);
		}
	}

	spin_unlock(&fdb->lock);
	rcu_read_unlock();

	work_delay = max_t(unsigned long, work_delay, msecs_to_jiffies(10));
	mod_delayed_work(system_long_wq, &fdb->gc_work, work_delay);
}

int __init awds_init(void)
{
	awds_fdb_cache = kmem_cache_create("awds_fdb_cache",
					   sizeof(struct awds_fdb_entry), 0,
					   SLAB_HWCACHE_ALIGN, NULL);
	if (!awds_fdb_cache)
		return -ENOMEM;

	return 0;
}

void __exit awds_exit(void)
{
	kmem_cache_destroy(awds_fdb_cache);
}
